# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass

from fastvideo.v1.configs.sample.base import SamplingParam


@dataclass
class StepVideoT2VSamplingParam(SamplingParam):
    # Video parameters
    height: int = 720
    width: int = 1280
    num_frames: int = 81

    # Denoising stage
    guidance_scale: float = 9.0
    num_inference_steps: int = 50

    # neg magic and pos magic
    # pos_magic: str = "超高清、HDR 视频、环境光、杜比全景声、画面稳定、流畅动作、逼真的细节、专业级构图、超现实主义、自然、生动、超细节、清晰。"
    # neg_magic: str = "画面暗、低分辨率、不良手、文本、缺少手指、多余的手指、裁剪、低质量、颗粒状、签名、水印、用户名、模糊。"
